package shopping.cart

import akka.actor.CoordinatedShutdown
import akka.actor.testkit.typed.scaladsl.{
  ActorTestKit,
  ScalaTestWithActorTestKit
}
import akka.actor.typed.ActorSystem
import akka.cluster.MemberStatus
import akka.cluster.typed.Cluster
import akka.grpc.GrpcClientSettings
import akka.testkit.SocketUtil
import com.typesafe.config.{ Config, ConfigFactory }
import org.scalatest.BeforeAndAfterAll
import org.scalatest.concurrent.{
  Eventually,
  PatienceConfiguration,
  ScalaFutures
}
import org.scalatest.matchers.should.Matchers
import org.scalatest.time.Span
import org.scalatest.wordspec.{ AnyWordSpec, AnyWordSpecLike }
import org.slf4j.LoggerFactory

import scala.concurrent.Future
import scala.concurrent.duration._

object IntegrationSpec {
  val sharedConfig: Config = ConfigFactory.load("integration-test.conf")

  private def nodeConfig(
      grpcPort: Int,
      managementPorts: Seq[Int],
      managementPortIndex: Int): Config =
    ConfigFactory.parseString(s"""
      shopping-cart-service.grpc {
        interface = "localhost"
        port = $grpcPort
      }
      akka.management.http.port = ${managementPorts(managementPortIndex)}
      akka.discovery.config.services {
        "shopping-cart-service" {
          endpoints = [
            {host = "127.0.0.1", port = ${managementPorts(0)}},
            {host = "127.0.0.1", port = ${managementPorts(1)}},
            {host = "127.0.0.1", port = ${managementPorts(2)}}
          ]
        }
      }
      """)

  class TestNodeFixture(
      grpcPort: Int,
      managementPorts: Seq[Int],
      managementPortIndex: Int) {
    val testKit =
      ActorTestKit(
        "IntegrationSpec",
        nodeConfig(grpcPort, managementPorts, managementPortIndex)
          .withFallback(sharedConfig)
          .resolve())

    def system: ActorSystem[_] = testKit.system

    private val clientSettings =
      GrpcClientSettings
        .connectToServiceAt("127.0.0.1", grpcPort)(testKit.system)
        .withTls(false)
    lazy val client: proto.ShoppingCartServiceClient =
      proto.ShoppingCartServiceClient(clientSettings)(testKit.system)
    CoordinatedShutdown
      .get(system)
      .addTask(
        CoordinatedShutdown.PhaseBeforeServiceUnbind,
        "close-test-client-for-grpc")(() => client.close());

  }
}

class IntegrationSpec
    extends AnyWordSpec
    with Matchers
    with BeforeAndAfterAll
    with ScalaFutures
    with Eventually {

  import IntegrationSpec.TestNodeFixture

  private val logger =
    LoggerFactory.getLogger(classOf[IntegrationSpec])

  implicit private val patience: PatienceConfig =
    PatienceConfig(10.seconds, Span(100, org.scalatest.time.Millis))

  private val (grpcPorts, managementPorts) =
    SocketUtil
      .temporaryServerAddresses(6, "127.0.0.1")
      .map(_.getPort)
      .splitAt(3)

  // one TestKit (ActorSystem) per cluster node
  private val testNode1 =
    new TestNodeFixture(grpcPorts(0), managementPorts, 0)
  private val testNode2 =
    new TestNodeFixture(grpcPorts(1), managementPorts, 1)
  private val testNode3 =
    new TestNodeFixture(grpcPorts(2), managementPorts, 2)

  private val systems3 =
    List(testNode1, testNode2, testNode3).map(_.testKit.system)

  override protected def beforeAll(): Unit = {
    super.beforeAll()
  }

  override protected def afterAll(): Unit = {
    super.afterAll()
    testNode3.testKit.shutdownTestKit()
    testNode2.testKit.shutdownTestKit()
    // testNode1 must be the last to shutdown
    // because responsible to close R2dbc connections
    testNode1.testKit.shutdownTestKit()
  }

  "Shopping Cart service" should {
    "init and join Cluster" in {
      Main.init(testNode1.testKit.system)
      Main.init(testNode2.testKit.system)
      Main.init(testNode3.testKit.system)

      // let the nodes join and become Up
      eventually(PatienceConfiguration.Timeout(15.seconds)) {
        systems3.foreach { sys =>
          Cluster(sys).selfMember.status should ===(MemberStatus.Up)
          Cluster(sys).state.members.unsorted.map(_.status) should ===(
            Set(MemberStatus.Up))
        }
      }

    }

    "update and project from different nodes via gRPC" in {
      // add from client1
      val response1 = testNode1.client.addItem(
        proto.AddItemRequest(cartId = "cart-1", itemId = "foo", quantity = 42))
      val updatedCart1 = response1.futureValue
      updatedCart1.items.head.itemId should ===("foo")
      updatedCart1.items.head.quantity should ===(42)

      // add from client2
      val response2 = testNode2.client.addItem(
        proto.AddItemRequest(cartId = "cart-2", itemId = "bar", quantity = 17))
      val updatedCart2 = response2.futureValue
      updatedCart2.items.head.itemId should ===("bar")
      updatedCart2.items.head.quantity should ===(17)

    }

  }
}
